import 'dart:io';
import 'dart:typed_data';
import 'http/http_connector.dart';
import 'moderator/OutgoingMessage.dart';
import 'moderator/OutgoingMessageCache.dart';
import 'moderator/moderatorCore.dart';
//import 'moderator/xxCache.dart';
import 'disk.dart';
export 'moderator/moderatorCore.dart';
export 'disk.dart';
import 'package:hex/hex.dart';
import 'xxhash/lib/fast_hash.dart';
import 'package:blurhash_dart/blurhash_dart.dart';
import 'package:image/image.dart' as img;
import 'ipfsQueue.dart';

class world {
  httpInterzone? _io;
  interzoneDisk? _disk;
  ModeratorEntrySet data = ModeratorEntrySet();
  ModeratorCapsule? _latestCapsule;
  Uint8List _katamari = Uint8List(0);
  xxCache _xxCache = xxCache();
  OutgoingMessageCache _ogMessages = OutgoingMessageCache();
  String _server = '';
  String _siteName = '';
  String get baseUrl => _server;
  String _nicke = '';
  String get currentNick => _nicke;
  String appSplashImage = '';
  bool _dirty = false;
  IpfsQueue _ipfsQ = IpfsQueue();

  void unDirty() {
    //call this from outside only
    _dirty = false;
  }

  bool get isDirty => _dirty;
  world();

  xxCache get xxC => _xxCache;
  //action handlers
  dynamic messageActionSignalHandler;
  Set<String> _mutedAuthors = {};
  Set<int> _deletedEntries = {};

  int get now => DateTime.now().millisecondsSinceEpoch;

  ModeratorEntrySet get messages {
    if (_mutedAuthors.isEmpty && _deletedEntries.isEmpty) return data;
    ModeratorEntrySet _l = ModeratorEntrySet();
    data.all.forEach((element) {
      if (_mutedAuthors.contains(element.author)) return;
      if (_deletedEntries.contains(element.keyHint)) return;
      _l.add(element);
    });
    return _l;
  }

  int xxHash(Uint8List b) {
    if (b.isEmpty) return -1;
    final digest = xxHash64.convert(b);
    return digest.x.toInt();
  }

  String xxHashHex(Uint8List data) {
    final t = xxHash(data);
    final bdata = Uint8List(8)..buffer.asByteData().setInt64(0, t, Endian.big);
    return HEX.encode(bdata);
  }

  List<ModeratorEntry>? updateLatestCapsule(Uint8List bin) {
    //throw latest shit in the entrySet
    print('updateLatestCapsule len');
    print(bin.lengthInBytes);
    var mc = ModeratorCapsule.fromBinary(bin);
    //if (mc.isDEADBEEF) return null;
    if (mc == null) return null;
    _latestCapsule = mc;
    _katamari = mc.katamari;
    //writes new shit to disk too
    final t = insertToLocalEntrySet(mc);
    if (t.isNotEmpty) {
      _dirty = true;
      _disk?.commitKatamari(mc.katamari);
    }
    return t;
  }

  Future<List<ModeratorEntry>?> init(String _siteName, String server,
      String author, interzoneDisk _disco) async {
    _server = server;
    _io = httpInterzone();
    _disk = _disco;
    //await _disk.init('interzoneData', false, _disk);
    String previousDartSeshId = '';
    // await _disk.readPreferences;

    List<String>? _prifs = await _disk?.readPreferences;
    if (_prifs != null) {
      previousDartSeshId = _prifs[0];
      if (_prifs.length > 1) {
        _nicke = _prifs[1];
      }
    } else {
      _nicke = author;
    }
    try {
      await _disk?.readDeletesPreferences.then((value) => {
            if (value != null)
              {
                if (value.isNotEmpty) {_deletedEntries = value.toSet()}
              }
          });
    } catch (e) {
      print('world::readDeletesPreferences exploded');
      print(e);
    }
    try {
      await _disk?.readMutesPreferences.then((value) => {
            if (value != null)
              {
                if (value.isNotEmpty)
                  {
                    _mutedAuthors =
                        value.where((element) => element.isNotEmpty).toSet()
                  }
              }
          });
    } catch (e) {
      print('world::readMutesPreferences exploded');
      print(e);
    }
    await _xxCache.init(_disk!);
    Uint8List? cps = await _disk!.readCapsule;
    if (cps != null) {
      LocalEntrySetPrime(cps);
    }
    Uint8List katt = await _disk!.readKatamari;
    if (katt.isNotEmpty) {
      _katamari = katt;
    }
    //async hundkarusellen!
    if (previousDartSeshId.isNotEmpty) {
      print('injected previous dart sesh id to help the police');
      _io!.dartSessionId = previousDartSeshId;
    }
    await _io!.handshake(_server, _katamari).then((hs) async {
      if (hs.isNotEmpty) {
        await _disk?.commitPreferences(_io!.dartSessionId, currentNick);
        satisfyXXwant(_io!.xxWantsList);
        final haa = ModeratorCapsule.fromBinary(hs);
        return updateLatestCapsule(hs);
      }
    });
    //we get moderator capsule

    return null;
  }

  Future<List<ModeratorEntry>?> refreshViaHTTP() async {
    if (_io == null) return null;
    return await _io!.refresh(_server, _katamari).then((hs) async {
      if (hs.isNotEmpty) {
        //await _disk.commitPreferences(_io!.dartSessionId, nick);

        satisfyXXwant(_io!.xxWantsList);
        return updateLatestCapsule(hs);
        outgoingMessageCacheSend();
      }
      return null;
    });
  }

  Future<bool> refreshTaggedViaHTTP(String tag) async {
    if (_io == null) return false;
    return await _io!.refreshTagged(_server, tag).then((hs) async {
      if (hs.isNotEmpty) {
        //await _disk.commitPreferences(_io!.dartSessionId, nick);
        updateLatestCapsule(hs);
        satisfyXXwant(_io!.xxWantsList);
        return true;
      }
      return false;
    });
  }

  Future<List<ModeratorEntry>?> refreshThreadViaHTTP(String shortlink) async {
    if (_io == null) return null;
    return await _io!.refreshThread(_server, shortlink).then((hs) async {
      if (hs.isNotEmpty) {
        //await _disk.commitPreferences(_io!.dartSessionId, nick);
        satisfyXXwant(_io!.xxWantsList);
        return updateLatestCapsule(hs);
      }
      return null;
    });
  }

  Future<bool> setUserNickAndImage(String _nick, Uint8List attachement) async {
    if (_nicke == _nick) return true;
    _nicke = _nick;

    await _disk?.commitPreferences(_io!.dartSessionId, _nick);
    return true;
  }

  Future<bool> newThreadToQueue(
      String _baseUrl,
      String text,
      Uint8List attachement,
      String _nick,
      String shortLinkOfPrevious,
      String cid) async {
    return _ogMessages.addPost(
        currentNick, text, shortLinkOfPrevious, '', now, attachement, cid);
  }

  Future<List<ModeratorEntry>?> newThread(
      String _baseUrl,
      String text,
      Uint8List attachement,
      String _nick,
      String shortLinkOfPrevious,
      String cid) async {
    if (_latestCapsule == null) return null;
    if (_nick.isEmpty) {
      _nick = currentNick;
    }

    //String pree =
    //    'SQvL0XBOR1I-WqlMMVolGlyFmIqc0qq4XnYOL8T7LgrGs5yYumNaRc9ZH-pLnN5g_RGrnCIYFDVvUD8ZQkYRcURfESYEEGYCOWcQESUdThwmAXgWOCAkCh4TFTctFnoJFSMIZyUWEAkLBUYCfScEAik6WS8LChMaUzwrACsaQXYJKjkkWDpO';
    Uint8List nt;
    if (attachement.isEmpty) {
      return _io!
          .newThread(_baseUrl, text, shortLinkOfPrevious,
              _latestCapsule!.katamari, currentNick)
          .then((value) {
        if (value.isEmpty) return null;
        return updateLatestCapsule(value);
      });
    }

    try {
      final image = img.decodeImage(attachement);
      final blurHash = BlurHash.encode(image!, numCompX: 4, numCompY: 3);
      text += ' !|' + blurHash.hash + ' ';
    } catch (e) {}

    if (cid.isNotEmpty) {
      return _io!
          .newThreadWithIPFSAttachement(
        _baseUrl,
        text,
        shortLinkOfPrevious,
        _latestCapsule!.katamari,
        currentNick,
        attachement.lengthInBytes,
        xxHash(attachement),
        cid,
      )
          .then((value) {
        if (value.isEmpty) return null;
        return updateLatestCapsule(value);
      });
    }

    print('attachemnt len ' + attachement.lengthInBytes.toString());
    cid = await pushToPinataIPFS(attachement);
    if (cid.isNotEmpty) {
      //moderator autosets attachmenet?
      text += ' !&' + cid + ' ';
    }
    return _io!
        .newThreadWithAttachement(
            _baseUrl,
            text,
            attachement,
            shortLinkOfPrevious,
            _latestCapsule!.katamari,
            currentNick,
            attachement.lengthInBytes,
            xxHash(attachement),
            cid)
        .then((value) {
      if (value.isEmpty) return null;
      _xxCache.pushXXbytes(attachement, false);
      return updateLatestCapsule(value);
    });
  }

  ModeratorCapsule? newGroup(String text, Uint8List attachement) {}
  ModeratorCapsule? newGroupChild(String text, Uint8List attachement) {}

  Future<List<ModeratorEntry>?> newReply(
      String text, String rootShortlink, String ppShortlink) async {
    Uint8List nt;
    nt = await _io!.newReply(text, rootShortlink, ppShortlink,
        _latestCapsule!.katamari, currentNick);

    if (nt.isEmpty) return null;
    return updateLatestCapsule(nt);
  }

  Future<bool> newReplyToQueue(
    String text,
    String shortLinkOfRoot,
    String shortLinkOfPP,
  ) async {
    return _ogMessages.addReply(currentNick, text, shortLinkOfRoot,
        shortLinkOfPP, now, Uint8List(0), '');
  }

  Future<Uint8List?> readLocalFileByFilename(String filename) async {
    String cp = Directory.current.path;
    try {
      var did = await File(Directory.current.path + '/testImages/' + filename)
          .readAsBytes();
      return did;
    } catch (e) {
      return null;
    }
  }

  List<ModeratorEntry> insertToLocalEntrySet(ModeratorCapsule c) {
    List<ModeratorEntry> _nw = [];
    final es = ModeratorEntrySet.fromModeratorCapsule(c.threadEntries);
    es.all.forEach((element) {
      print(element.hint);
      print(element.bl);
      print(element.body);
    });
    if (es.all.isEmpty) return _nw;
    var i = 0;
    es.all.forEach((element) {
      if (_mutedAuthors.contains(element.author)) {
        return;
      }
      if (_deletedEntries.contains(element.keyHint)) {
        return;
      }

      if (data.add(element)) {
        _nw.add(element);
        i++;
      }
    });
    if (i == 0) return _nw; //no new entries

    Uint8List ul = ModeratorCapsuleHelper(data);
    if (ul.isNotEmpty) {
      _disk?.commitCapsule(ul);
    }
    return _nw;
  }

  LocalEntrySetPrime(Uint8List l) {
    final es = ModeratorEntrySet.fromModeratorCapsule(l);
    if (es.all.isEmpty) return;
    var i = 0;
    for (var ii in es.all) {
      if (_mutedAuthors.contains(ii.author)) return;
      if (_deletedEntries.contains(ii.hint)) return;
      if (data.add(ii)) {
        i++;
      }
    }
    if (i > 0) _dirty = true;
    print('LocalEntrySetPrime with ' + i.toString());
  }

  Uint8List ModeratorCapsuleHelper(ModeratorEntrySet mode) {
    List<ModeratorCapsuleItem> mcl = [];

    if (mode.all.isEmpty) {
      print('PANIC: moderator capsule helper, moderator entryset is empty');
      return Uint8List(0);
    }

    //pluggy backlinks

    //pluggy perpre

    //mangle

    for (var ee in mode.all) {
      mcl.add(ModeratorCapsuleItem(
          ee.flags.binary,
          0,
          ee.bl,
          ee.pp,
          ee.timeElapsedSincePostMs,
          ee.body.length,
          ee.bodyBytes,
          ee.shortLinkBytes.lengthInBytes,
          ee.shortLinkBytes,
          ee.keyHint,
          ee.avatar,
          ee.nick,
          ee.attachmentLink,
          0,
          ee.perceivedTimestamp,
          ee.likes,
          ee.shares));
    }

    var fppf = mcl.last;
    var fppfb = mcl.last.binary;
    ModeratorCapsuleItem popo = ModeratorCapsuleItem.fromBinary(fppfb);

    Uint8List enl = Uint8List(1064000);

    var wi = 0;
    for (var element in mcl) {
      if (wi < 3064000 / 0.8) {
        element.binary.forEach((element2) {
          enl[wi++] = element2;
        });
      }
    }

    return enl.sublist(0, wi);
  }

  List<List<dynamic>> _xxPullQueue = [];
  Set<int> _xxCheckedQueue = {};
  int _xxpqcou = 0;

  scanForCachedAttachements() async {
    //proto pull all attachemendos
    _xxpqcou = 0;

    if (_xxCheckedQueue.length < _xxPullQueue.length) return;
    if (data.all.isNotEmpty) {
      for (var element in data.all) {
        if (element.flags.attachements) {
          if (_xxCache.isXXcached(element.attachmentLinkAsXXint)) return;
          //print(element.shortLink);
          _xxPullQueue.add([element.attachmentLinkAsXXint, element.shortLink]);
        }
      }
      ;
    }
    if (_xxPullQueue.isEmpty) return;
  }

  primeCurrentViewWithCachedAttachments(ModeratorEntrySet m, int amount) {
    for (var ii = 0; ii < m.all.length; ii++) {
      if (!data.all[ii].flags.attachements) continue;
      if (_xxCache.isXXcached(m.all[ii].attachmentLinkAsXXint)) continue;
      _xxPullQueue.add([m.all[ii].attachmentLinkAsXXint, m.all[ii].shortLink]);
      amount--;
      if (amount < 1) break;
    }
  }

  pullForCachedAttachements() async {
    //if (_xxCheckedQueue.length == _xxPullQueue.length) return;
    if (_xxPullQueue.isEmpty) return;
    /*if (_xxpqcou > _xxPullQueue.length - 1) {
      _xxpqcou = 0;
      return;
    }*/
    //if (_xxpqcou > _xxPullQueue.length) return;
    Uint8List? xx = await _xxCache
        .pullXX(_xxPullQueue[_xxpqcou][0], _xxPullQueue[_xxpqcou][1], _server)
        .then((value) => null);
    _xxCheckedQueue.add(_xxpqcou);
    _xxpqcou++;
    if (xx == null) return;
    print(xx.lengthInBytes);
    _xxCheckedQueue.add(_xxpqcou);
    //if (_xxpqcou < _xxPullQueue.length - 1) {
    _xxpqcou++;
    //} else {
    //  _xxpqcou = 0;
    //}
  }

  void satisfyXXwant(Uint8List _wants) async {
    if (_wants.isEmpty) return;
    var s = 0;
    var b = new ByteData.view(_wants.buffer);
    var w = 0;
    while (w < _wants.lengthInBytes) {
      var xx = b.getUint64(w);
      if (_xxCache.isXXcached(xx)) {
        s = xx;
        break;
      }
      w += 8;
    }
    if (s == 0) return;
    Uint8List? data = await _xxCache.pullXX(s, '', '');
    if (data != null) {
      _io!.satisfyXXwant(data);
    }
  }

  Future<String> pushToPinataIPFS(Uint8List data) async {
    //add your pinata.cloud keys to PinataApiKeys.dart
    return _ipfsQ.insertToQueue(xxHash(data), data);
    //return _io!.pushToPinataIPFS(pinata_api_key, pinata_secret_api_key, data);
  }

  Future<bool> deleteEntry(ModeratorEntry m) async {
    if (_deletedEntries.contains(m.keyHint)) return false;
    _deletedEntries.add(m.keyHint);
    //TODO kick off local storage too
    //xxC.delete(m.attachmentLink)
    //if it's your / your group, kill from serv too
    //await _disk?.commitPreferences(_io!.dartSessionId, _nick);
    await _disk?.commitDeletesPreferences(_deletedEntries.toList());
    if (m.flags.moderate) {
      _ogMessages.addDrop(m.shortLink, now);
    }
    _dirty = true;
    return true;
  }

  Future<bool> muteAuthor(ModeratorEntry m) async {
    //if (m.flags.moderate) return false;
    if (_mutedAuthors.contains(m.keyHint)) return false;
    _mutedAuthors.add(m.author);

    await _disk?.commitMutesPreferences(_mutedAuthors.toList());

    _ogMessages.addMute(m.shortLink, now);
    _dirty = true;
    return true;
  }

  Future<bool> shareEntry(ModeratorEntry m) async {
    if (m.flags.moderate) return false;
    _ogMessages.addShare(m.shortLink, now);
    return true;
  }

  bool _processingoutgoingMessageCacheSend = false;

  Future<List<ModeratorEntry>?> outgoingMessageCacheSend() async {
    if (_processingoutgoingMessageCacheSend) return null;
    final pi = _ogMessages.pendingItem;
    if (pi == null) return null;
    _processingoutgoingMessageCacheSend = true;

    //if the request does not work, this prevents blocking the queue
    _ogMessages.holdPendingItem(pi.timestamp);

    if (pi.type == 3) {
      final r = await newDrop(pi.shortLinkOfPrevious, pi.shortLinkOfPP);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        return r;
      }
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    if (pi.type == 4) {
      final r = await newMute(pi.shortLinkOfPrevious);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        return r;
      }
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    if (pi.type == 5) {
      final r = await newShare(pi.shortLinkOfPrevious);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        return r;
      }
      _ogMessages.releasePendingItem(pi.timestamp);
    }

    if (pi.type == 1) {
      final r =
          await newReply(pi.text, pi.shortLinkOfPrevious, pi.shortLinkOfPP);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        return r;
      }
      _ogMessages.releasePendingItem(pi.timestamp);
    }

    if (pi.type == 0 && pi.attachement.isEmpty) {
      final r = await newThread(baseUrl, pi.text, pi.attachement, pi.nick,
          pi.shortLinkOfPrevious, pi.cid);
      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        return r;
      }
      _ogMessages.releasePendingItem(pi.timestamp);
    }

    if (pi.type == 0 && pi.attachement.isNotEmpty && pi.cid.isEmpty) {
      //inject pinata keys here
      final cid = await pushToPinataIPFS(pi.attachement);
      if (cid.isNotEmpty) {
        _xxCache.pushXXbytes(pi.attachement, false);
        final r = await newThread(baseUrl, pi.text, pi.attachement, pi.nick,
            pi.shortLinkOfPrevious, cid);
        if (r != null) {
          _ogMessages.commitSent(pi.timestamp);
          _processingoutgoingMessageCacheSend = false;
          return r;
        }
        _ogMessages.releasePendingItem(pi.timestamp);
      }
    }
    _processingoutgoingMessageCacheSend = false;
  }

  Future<List<ModeratorEntry>?> newDrop(
      String rootShortlink, String ppShortlink) async {
    Uint8List nt;
    nt = await _io!.newDrop(rootShortlink, _latestCapsule!.katamari);

    if (nt.isEmpty) return null;
    return updateLatestCapsule(nt);
  }

  Future<List<ModeratorEntry>?> newMute(String rootShortlink) async {
    Uint8List nt;
    nt = await _io!.newMute(rootShortlink, _latestCapsule!.katamari);

    if (nt.isEmpty) return null;
    return updateLatestCapsule(nt);
  }

  Future<List<ModeratorEntry>?> newLike(String rootShortlink) async {
    Uint8List nt;
    nt = await _io!.newLike(rootShortlink, _latestCapsule!.katamari);

    if (nt.isEmpty) return null;
    return updateLatestCapsule(nt);
  }

  Future<List<ModeratorEntry>?> newShare(String rootShortlink) async {
    Uint8List nt;
    nt = await _io!.newShare(rootShortlink, _latestCapsule!.katamari);

    if (nt.isEmpty) return null;
    return updateLatestCapsule(nt);
  }
}
